--- /usr/src/sys/i386/i386/machdep.c	Tue Feb  6 10:30:00 2007
+++ machdep.c	Thu Feb 22 04:25:35 2007
@@ -88,6 +88,7 @@ __FBSDID("$FreeBSD: src/sys/i386/i386/ma
 #include <sys/ucontext.h>
 #include <sys/vmmeter.h>
 
+#include <vm/uma.h>
 #include <vm/vm.h>
 #include <vm/vm_extern.h>
 #include <vm/vm_kern.h>
@@ -163,6 +164,7 @@ extern void initializecpu(void);
 static void cpu_startup(void *);
 static void fpstate_drop(struct thread *td);
 static void get_fpcontext(struct thread *td, mcontext_t *mcp);
+static int get_align(void);
 static int  set_fpcontext(struct thread *td, const mcontext_t *mcp);
 #ifdef CPU_ENABLE_SSE
 static void set_fpregs_xmm(struct save87 *, struct savexmm *);
@@ -2049,7 +2051,7 @@ init386(first)
 	int first;
 {
 	struct gate_descriptor *gdp;
-	int gsel_tss, metadata_missing, x;
+	int gsel_tss, metadata_missing, x, ualign;
 	struct pcpu *pc;
 
 	thread0.td_kstack = proc0kstack;
@@ -2301,6 +2303,9 @@ init386(first)
 	_ucodesel = GSEL(GUCODE_SEL, SEL_UPL);
 	_udatasel = GSEL(GUDATA_SEL, SEL_UPL);
 
+	if ((ualign = get_align()) != -1)
+		uma_set_align(ualign);
+
 	/* setup proc 0's pcb */
 	thread0.td_pcb->pcb_flags = 0;
 #ifdef PAE
@@ -2672,6 +2677,85 @@ get_fpcontext(struct thread *td, mcontex
 	}
 	mcp->mc_fpformat = npxformat();
 #endif
+}
+
+static int
+get_align(void)
+{
+	unsigned char buff[16], again;
+	int lone, ltwo;
+	unsigned short cnt;
+
+	bzero(buff, 16 * sizeof(unsigned char));
+	lone = 0;
+	ltwo = 0;
+
+	do {
+		__asm __volatile (
+		"xorl %%eax,%%eax ;		"
+		"movb $2,%%al ;			"
+		"cpuid ;			"
+		"movl %1,%%edi ;		"
+		"movb %%al,%0 ;			"
+		"testl $0x80000000,%%eax ;	"
+		"jnz 1f ;			"
+		"movl %%eax,(%%edi) ;		"
+		"1: ;				"
+		"testl $0x80000000,%%ebx ;	"
+		"jnz 2f ;			"
+		"movl %%ebx,4(%%edi) ;		"
+		"2: ;				"
+		"testl $0x80000000,%%ecx ;	"
+		"jnz 3f ;			"
+		"movl %%ecx,8(%%edi) ;		"
+		"3: ;				"
+		"testl $0x80000000,%%edx ;	"
+		"jnz 4f ;			"
+		"movl %%edx,12(%%edi) ;		"
+		"4: ;				"
+		: "=g"(again),
+		: "g"(buff)
+		: "%eax", "%ebx", "%ecx", "%edx", "%edi", "memory");
+
+		for (cnt = 0; cnt < 16; cnt++) {
+			switch(buff[cnt]) {
+			case 0x0A:
+			case 0x0C:
+				lone = 32 - 1;
+				break;
+			case 0x2C:
+			case 0x66:
+			case 0x67:
+			case 0x68:
+				lone = 64 - 1;
+				break;
+			case 0x41:
+			case 0x42:
+			case 0x43:
+			case 0x44:
+			case 0x45:
+			case 0x82:
+			case 0x83:
+			case 0x84:
+			case 0x85:
+				ltwo = 32 - 1;
+				break;
+			case 0x78:
+			case 0x79:
+			case 0x7A:
+			case 0x7B:
+			case 0x7C:
+			case 0x86:
+			case 0x87:
+				ltwo = 64 - 1;
+				break;
+			}
+			if (lone && ltwo)
+				break;
+		}
+	} while (again > 1);
+
+	return ((!lone || !ltwo) ? -1 : MAX(lone, ltwo));
 }
 
 static int
